@@ -6,11 +6,12 @@ Parallel::ForkManager - A simple parallel processing fork manager
 
   use Parallel::ForkManager;
 
-  $pm = Parallel::ForkManager->new($MAX_PROCESSES);
+  my $pm = Parallel::ForkManager->new($MAX_PROCESSES);
 
-  foreach $data (@all_data) {
+  DATA_LOOP:
+  foreach my $data (@all_data) {
     # Forks and returns the pid for the child:
-    my $pid = $pm->start and next;
+    my $pid = $pm->start and next DATA_LOOP;
 
     ... do some work with $data in the child process ...
 
@@ -30,7 +31,7 @@ The code for a downloader would look something like this:
 
   ...
 
-  @links=(
+  my @links=(
     ["http://www.foo.bar/rulez.data","rulez_data.txt"],
     ["http://new.host/more_data.doc","more_data.doc"],
     ...
@@ -41,12 +42,13 @@ The code for a downloader would look something like this:
   # Max 30 processes for parallel download
   my $pm = Parallel::ForkManager->new(30);
 
+  LINKS:
   foreach my $linkarray (@links) {
-    $pm->start and next; # do the fork
+    $pm->start and next LINKS; # do the fork
 
-    my ($link,$fn) = @$linkarray;
+    my ($link, $fn) = @$linkarray;
     warn "Cannot get $fn from $link"
-      if getstore($link,$fn) != RC_OK;
+      if getstore($link, $fn) != RC_OK;
 
     $pm->finish; # do the exit in the child process
   }
@@ -119,6 +121,27 @@ Allows you to set a new maximum number of children to maintain.
 You can call this method to wait for all the processes which have been
 forked. This is a blocking wait.
 
+=item max_procs 
+
+Returns the maximal number of processes the object will fork.
+
+=item running_procs
+
+Returns the pids of the forked processes currently monitored by the
+C<Parallel::ForkManager>. Note that children are still reported as running
+until the fork manager harvest them, via the next call to
+C<start> or C<wait_all_children>.
+
+    my @pids = $pm->running_procs;
+
+    my $nbr_children =- $pm->running_procs;
+
+=item wait_for_available_procs( $n )
+
+Wait until C<$n> available process slots are available.
+If C<$n> is not given, defaults to I<1>.
+
+
 =back
 
 =head1 CALLBACKS
@@ -209,16 +232,19 @@ This small example can be used to get URLs in parallel.
 
   use Parallel::ForkManager;
   use LWP::Simple;
-  my $pm= Parallel::ForkManager->new(10);
+
+  my $pm = Parallel::ForkManager->new(10);
+
+  LINKS:
   for my $link (@ARGV) {
-    $pm->start and next;
-    my ($fn)= $link =~ /^.*\/(.*?)$/;
+    $pm->start and next LINKS;
+    my ($fn) = $link =~ /^.*\/(.*?)$/;
     if (!$fn) {
       warn "Cannot determine filename from $fn\n";
     } else {
-      $0.=" ".$fn;
+      $0 .= " " . $fn;
       print "Getting $fn from $link\n";
-      my $rc=getstore($link,$fn);
+      my $rc = getstore($link, $fn);
       print "$link downloaded. response code: $rc\n";
     };
     $pm->finish;
@@ -246,7 +272,7 @@ Example of a program using callbacks to get child exit codes:
   });
 
   $pm->run_on_start( sub {
-      my ($pid,$ident)=@_;
+      my ($pid, $ident)=@_;
       print "** $ident started, pid: $pid\n";
   });
 
@@ -256,8 +282,9 @@ Example of a program using callbacks to get child exit codes:
     0.5
   );
 
+  NAMES:
   foreach my $child ( 0 .. $#names ) {
-    my $pid = $pm->start($names[$child]) and next;
+    my $pid = $pm->start($names[$child]) and next NAMES;
 
     # This code is the child process
     print "This is $names[$child], Child number $child\n";
@@ -289,7 +316,8 @@ In this simple example, each child sends back a string reference.
       if (defined($data_structure_reference)) {  # children are not forced to send anything
         my $string = ${$data_structure_reference};  # child passed a string reference
         print "$string\n";
-      } else {  # problems occuring during storage or retrieval will throw a warning
+      }
+      else {  # problems occuring during storage or retrieval will throw a warning
         print qq|No message received from child process $pid!\n|;
       }
     }
@@ -300,9 +328,9 @@ In this simple example, each child sends back a string reference.
   my @preferences = ('loves', q|can't stand|, 'always wants more', 'will walk 100 miles for', 'only eats', 'would starve rather than eat');
 
   # run the parallel processes
-  my $person = '';
-  foreach $person (qw(Fred Wilma Ernie Bert Lucy Ethel Curly Moe Larry)) {
-    $pm->start() and next;
+  PERSONS:
+  foreach my $person (qw(Fred Wilma Ernie Bert Lucy Ethel Curly Moe Larry)) {
+    $pm->start() and next PERSONS;
 
     # generate a random statement about food preferences
     my $statement = $person . ' ' . $preferences[int(rand @preferences)] . ' ' . $foods[int(rand @foods)];
@@ -356,9 +384,9 @@ process whatever is retrieved.
     {'name' => 'ENV values in an array', 'send' => 'values'},
   );
 
-  my $instruction = '';
-  foreach $instruction (@instructions) {
-    $pm->start($instruction->{'name'}) and next;  # this time we are using an explicit, unique child process identifier
+  INSTRUCTS:
+  foreach my $instruction (@instructions) {
+    $pm->start($instruction->{'name'}) and next INSTRUCTS;  # this time we are using an explicit, unique child process identifier
 
     # last step in child processing
     $pm->finish(0) unless $instruction->{'send'};  # no data structure is sent unless this child is told what to send.
@@ -424,7 +452,7 @@ and/or modify it under the same terms as Perl itself.
 
 =head1 CREDITS
 
-  Gábor Szabó (szabgab@cpn.org)  (co-maintainer)
+  Gábor Szabó (szabgab@cpan.org)  (co-maintainer)
   Michael Gang (bug report)
   Noah Robin <sitz@onastick.net> (documentation tweaks)
   Chuck Hirstius <chirstius@megapathdsl.net> (callback exit status, example)
@@ -440,9 +468,11 @@ use Storable qw(store retrieve);
 use File::Spec;
 use File::Temp ();
 use File::Path ();
+use Carp;
+
 use strict;
 use vars qw($VERSION);
-$VERSION="1.04";
+$VERSION="1.08";
 $VERSION = eval $VERSION;
 
 sub new {
@@ -453,6 +483,7 @@ sub new {
     processes  => {},
     in_child   => 0,
     parent_pid => $$,
+    auto_cleanup => ($tempdir ? 0 : 1),
   };
 
 
@@ -572,6 +603,25 @@ sub wait_all_children {
 
 *wait_all_childs=*wait_all_children; # compatibility;
 
+sub max_procs { $_[0]->{max_proc}; }
+
+sub running_procs {
+    my $self = shift;
+
+    my @pids = keys %{ $self->{processes} };
+    return @pids;
+}
+
+sub wait_for_available_procs {
+    my( $self, $nbr ) = @_;
+    $nbr ||= 1;
+
+    croak "nbr processes '$nbr' higher than the max nbr of processes (@{[ $self->max_procs ]})"
+        if $nbr > $self->max_procs;
+
+    $self->wait_one_child until $self->max_procs - $self->running_procs >= $nbr;
+}
+
 sub run_on_finish {
   my ($s,$code,$pid)=@_;
 
@@ -622,44 +672,45 @@ sub set_max_procs {
   $s->{max_proc} = $mp;
 }
 
-# OS dependant code follows...
-
 sub _waitpid { # Call waitpid() in the standard Unix fashion.
-  return waitpid($_[1],$_[2]);
+    my( $self, $pid, $flag ) = @_;
+
+    return $flag ? $self->_waitpid_non_blocking : $self->_waitpid_blocking;
 }
 
-# On ActiveState Perl 5.6/Win32 build 625, waitpid(-1, &WNOHANG) always
-# blocks unless an actual PID other than -1 is given.
-sub _NT_waitpid {
-  my ($s, $pid, $par) = @_;
+sub _waitpid_non_blocking {
+    my $self = shift;
 
-  if ($par == &WNOHANG) { # Need to nonblock on each of our PIDs in the pool.
-    my @pids = keys %{ $s->{processes} };
-    # Simulate -1 (no processes awaiting cleanup.)
-    return -1 unless scalar(@pids);
-    # Check each PID in the pool.
-    my $kid;
-    foreach $pid (@pids) {
-      $kid = waitpid($pid, $par);
-      return $kid if $kid != 0; # AS 5.6/Win32 returns negative PIDs.
+    for my $pid ( $self->running_procs ) {
+        my $p = waitpid $pid, &WNOHANG or next;
+        if ( $p == -1 ) {
+            warn "child process '$pid' disappeared. A call to `waitpid` outside of Parallel::ForkManager might have reaped it.\n";
+            # it's gone. let's clean the process entry
+            delete $self->{processes}{$pid};
+        }
+        else {
+            return $pid;
+        }
     }
-    return $kid;
-  } else { # Normal waitpid() call.
-    return waitpid($pid, $par);
-  }
+
+    return 0;
 }
 
-{
-  local $^W = 0;
-  if ($^O eq 'NT' or $^O eq 'MSWin32') {
-    *_waitpid = \&_NT_waitpid;
-  }
+sub _waitpid_blocking {
+    my $self = shift;
+
+    while() {
+        my $pid = $self->_waitpid_non_blocking;
+        return $pid if $pid;
+
+        sleep 1;
+    }
 }
 
 sub DESTROY {
   my ($self) = @_;
 
-  if ($self->{parent_pid} == $$ && -d $self->{tempdir}) {
+  if ($self->{auto_cleanup} && $self->{parent_pid} == $$ && -d $self->{tempdir}) {
     File::Path::remove_tree($self->{tempdir});
   }
 }
